# niconico.py - Client

from __future__ import annotations

from typing import Optional

from logging import getLogger
import re

import requests

from .video import Client as VideoClient
from .search import SearchClient
from .cookies import Cookies

from .exceptions import LoginFailureException


__all__ = ("adjust_cookies", "NicoNico", "logger")
Response = requests.Response
logger = getLogger("niconico.py")
"``logging`` の ``Logger`` です。"


def adjust_cookies(cookies: Cookies) -> dict[str, str]:
    """Cookiesを辞書に変換します。

    Parameters
    ----------
    cookies : Cookies
        辞書に変換するクッキーです。"""
    return {key: morsel.value for key, morsel in cookies.items()}


class NicoNico:
    """ニコニコにアクセスするために使うクラスです。

    Parameters
    ----------
    cookies : Cookies, optional
        リクエストの際に使用するクッキーです。"""

    video: VideoClient
    "ニコニコ動画用のクライアントクラスのインスタンスです。"
    cookies: Optional[Cookies]
    "リクエスト時に使用するクッキーです。"

    def __init__(self, cookies: Optional[Cookies] = None):
        self.video = VideoClient(self)
        self.searchclient = SearchClient(self)
        self.cookies = cookies
        self.logger = logger

    def request(self, method: str, url: str, *args, **kwargs) -> requests.Response:
        """``requests.request`` を使用して設定されているクッキーでリクエストを行います。
        引数は ``requests.request`` と同じです。

        Notes
        -----
        引数 ``cookies`` に :class:`niconico.cookies.Cookies` を渡した場合は自動で辞書に変換されます。"""
        save_default = True

        # クッキーの調整をする。
        if "cookies" not in kwargs:
            save_default = False
            if self.cookies is not None:
                # kwargsにクッキーがないのならそのクッキーを設定する。
                kwargs["cookies"] = self.cookies
            elif kwargs.pop("require_cookies", False):
                # もしクッキーが必須の場合かつクッキーが指定されていないのならゲスト用のクッキーを作る。
                kwargs["cookies"] = Cookies.guest()
        # クッキーがCookiesクラスになっているなら辞書にする。
        if isinstance(kwargs.get("cookies", {}), Cookies):
            kwargs["cookies"] = adjust_cookies(kwargs["cookies"])

        # デバッグログを出力する。
        self.logger.debug("リクエスト: (%s) %s" % (method, url))
        for key, value in (
            ("cookies", "クッキー"), ("params", "パラメーター"),
            ("json", "JSON")
        ):
            if key in kwargs:
                self.logger.debug("使用予定の%s: %s" % (value, kwargs[key]))

        # リクエストをする。
        response = requests.request(method, url, *args, **kwargs)
        if save_default:
            self.cookies = Cookies(response.cookies.get_dict())
        response.raise_for_status()
        return response

    def login(self, mail: str, password: str) -> NicoNico:
        """メールアドレスとパスワードを用いてログインを行います。
        二段階認証が有効になっているアカウントではログインすることができません。
        クッキーの中身を直接置き換える方法で認証をしてください。

        Parameters
        ----------
        mail : str
            ログインする際のメールアドレスもしくは電話番号です。
        password : str
            ログインする際のパスワードです。

        Raises
        ------
        LoginFailureException"""
        session = requests.session()
        
        res = session.post(
            "https://secure.nicovideo.jp/secure/login?site=niconico",
            params={
                "mail": mail,
                "password": password
            })
        
        if res.headers.get("x-niconico-authflag") == ('1' or '3'):
            self.cookies = Cookies.from_string(session.cookies.get("user_session"))
            return self
        else:
            title_ptn = re.compile('<title>(.*?)</title>')
            title = title_ptn.search(res.text)
            if title:
                if "2段階認証" in title.group(1):
                    raise LoginFailureException("Two-step verification is not supported.")
        raise LoginFailureException("Login failed.")



# niconico.py - Comment

from typing import TYPE_CHECKING, TypedDict, Literal, Optional, List, Any
from niconico.base import DictFromAttribute


__all__ = (
    "Comments", "MovieChat"
)


class MovieChat(DictFromAttribute):
    "コメントデータです。"

    thread: str
    "スレッドのIDです。"
    no: int
    "コメント番号です。"
    vpos: int
    "コメントが投稿された場所です。"
    date: int
    "コメントが投稿された日時です。"
    nicoru: int = 0
    "ニコられた数です。"
    premium: bool = False
    "プレミアムかどうかです。"
    anonymity: bool
    "匿名アカウントかどうかです。"
    date_usec: int = 0
    "コメントが投稿された日時のミリ秒数です。"
    score: int = 0
    "コメントのNG共有スコアの値です。"
    user_id: str
    "コメント投稿者のユーザーIDです。"
    mail: str
    "指定されたコマンドです。"
    content: str
    "コメント内容です。"

    def __init__(self, data: dict):
        self.__data__ = data
        super().__init__(self.__data__, self)


class Comments():
    "動画のコメントデータです。"

    thread: str
    "スレッドのIDです。"
    fork: str
    "コメントの種類です。"
    count: int
    "投稿されたコメント数です。"
    last_res: Optional[int] = 0
    "取得したものの最新のコメント番号です。"
    num_res: int
    "全サーバー合計のコメント数です。"
    ticket: Optional[str]
    "コメント投稿をする際のチケットです。"
    chats: List[MovieChat]
    "コメントデータのリストです。"
    when: int
    "いつ時点のコメントデータかをUNIXタイムで返します。"

    def __init__(self):
        self.chats = []


# niconico.py - Cookies

from typing import Optional

from http.cookies import SimpleCookie

from datetime import datetime, timedelta
from time import time


__all__ = ("Cookies",)
FORMAT = "%a, %d-%b-%Y %X"


class Cookies(SimpleCookie):
    @staticmethod
    def _make_cookie_tuple(name: str, cookie: str, domain: Optional[str] = ".nicovideo.jp"):
        cookies = Cookies()
        cookies[name] = cookie
        for key, value in (
            ("domain", domain), ("path", "/"),
            ("expires", (datetime.now() + timedelta(days=365)).strftime(FORMAT))
        ):
            cookies[name][key] = value
        return cookies

    @classmethod
    def from_file(cls, path: str):
        """Netscapeのクッキーファイルフォーマットに準拠したテキストファイルからクッキーが格納されたクラスを作成します。  
        なので、ChromeやEdgeで使用可能な `Get cookies.txt <https://chrome.google.com/webstore/detail/get-cookiestxt/bgaddhkoddajcdgocldbbfleckgcbcid>`_ のような拡張機能を使って作ったクッキーのテキストファイルを読み込むことができます。
        もし自分のアカウントでニコニコ動画のコンテンツを取得したい場合はこれを使用しましょう。

        Parameters
        ----------
        path : str
            クッキーのファイルのパスです。

        Raises
        ------
        FileNotFoundError"""
        with open(path, "r") as f:
            raw = f.read()
        cookies = cls()
        for item in map(
            lambda x: x.split(), filter(
                lambda x: x and x[0] != "#", raw.splitlines()
            )
        ):
            cookies[item[5]] = item[6]
            for index, key in enumerate(
                ("domain", None, "path", "secure", "expires")
            ):
                if key:
                    cookies[item[5]][key] = (
                        datetime.fromtimestamp(float(item[index])).strftime(FORMAT)
                        if key == "expires" else item[index]
                    )
        return cookies
    
    @classmethod
    def from_string(cls, user_session: str):
        """ニコニコ動画上での認証済みのクッキーの値を直接指定しクラスを作成します。

        Parameters
        ----------
        user_session : str
            ユーザーセッションです。"""
        return cls._make_cookie_tuple("user_session", user_session)

    @classmethod
    def guest(cls, nicosid: Optional[str] = None):
        """ニコニコのゲストアカウントのクッキーを生成します。  

        Notes
        -----
        これはニコニコ動画にアクセスして作成するクッキーではなく、開発者がクッキーを見て予想して作った再現物です。"""
        return cls._make_cookie_tuple("nicosid", nicosid or str(time()))


# niconico.py - Video # TODO: middleが何なのか調べる。

from __future__ import annotations

from typing import TYPE_CHECKING, TypedDict, Literal, Optional, List, Any

from ..base import DictFromAttribute

if TYPE_CHECKING:
    from ..video import Client as VideoClient, Video as RealVideo


__all__ = (
    "NicoDic", "Phrase", "EasyComment", "TagEdit", "TagItem", "Tag", "Counter", "Thumbnail",
    "Rating", "ViewerLike", "Viewer", "AbcOwner", "AbcVideo", "Video", "VideoOwner",
    "CommentServer","NvCommentID","NvCommentParams","NvComment","Comment",
    "MyListOwner", "MyListItemVideo", "MyListItem", "MyList"
)


# easyComment
class NicoDic(DictFromAttribute):
    "ニコニコ大百科での簡易データです。"

    title: str
    viewTitle: str
    summary: str
    link: str


class Phrase(DictFromAttribute):
    "言葉です。"

    text: str
    "言葉です。"
    nicodic: NicoDic
    "ニコニコ大百科のデータです。"


class EasyComment(DictFromAttribute):
    "簡易コメントリストです。"

    phrases: List[Phrase]
    "簡易コメントリストのデータです。"


# tag
class TagEdit(DictFromAttribute):
    "タグを編集可能かどうかのデータです。"

    isEditable: bool
    "編集可能かどうかです。"
    uneditableReason: Optional[Literal["NEED_LOGIN"]]
    "タグを編集できない理由です。"
    editKey: Optional[str]
    "不明"


class TagItem(DictFromAttribute):
    "タグの内容です。"

    name: str
    "タグの名前"
    isCategory: bool
    "不明"
    isCategoryCandidate: bool
    "不明"
    isNicodicArticleExists: bool
    "不明"
    isLocked: bool
    "ロックされているかどうかです。"


class Tag(DictFromAttribute):
    "タグデータです。"

    items: List[TagItem]
    "ついているタグのリストです。"
    hasR18Tag: bool
    "センシティブなタグがあるかどうかです。"
    isPublishedNicoscript: bool
    "不明"
    edit: TagEdit
    "編集できるかのデータです。"
    viewer: Optional[TagEdit]
    "不明"


# video
class Counter(DictFromAttribute):
    "動画の閲覧数等のカウンターです。"

    view: int
    "閲覧数"
    comment: int
    "コメント数"
    mylist: int
    "マイリスト数"
    like: int
    "「好き」数"


class Thumbnail(DictFromAttribute):
    "サムネイルです。"

    url: str
    "サムネイルの画像のURLです。"
    middleUrl: str
    "不明"
    largeUrl: str
    "不明"
    player: str
    "不明"
    ogp: str
    "不明"


class Rating(DictFromAttribute):
    "動画の評価です。"

    isAdult: bool
    "大人向けかどうかです。"


class ViewerLike(DictFromAttribute):
    "クライアントの動画へやった「好き」のデータです。"

    isLiked: bool
    "「好き」を押したかどうかです。"
    count: Any
    "不明"


class Viewer(DictFromAttribute):
    "動画でのクライアントの内容です。"

    isOwner: bool
    "動画投稿者かどうかです。"
    like: ViewerLike
    "動画へやった「好き」のデータです。"


class AbcOwner(DictFromAttribute):
    "動画投稿者の基底クラスです。"

    id: str
    "投稿者のIDです。"
    name: str
    "投稿者の名前です。"
    iconUrl: str
    "投稿者のアイコンのURLです。"


class AbcVideo(DictFromAttribute["VideoClient"]):
    "動画データの基底クラスです。"

    id: str
    "動画IDです。"
    title: str
    "動画のタイトルです。"
    description: str
    "動画の説明です。"
    count: Counter
    "動画の閲覧数等のデータです。"
    duration: int
    "動画の長さです。"
    thumbnail: Thumbnail
    "動画のサムネイルです。"
    registeredAt: str
    "動画のアップロードされた時間です。"


class VideoOwner(AbcOwner):
    "動画の投稿者のデータです。"

    isVideosPublic: bool
    "アップロードした動画が公開されているかどうかです。"
    isMylistsPublic: bool
    "マイリストが公開されているかどうかです。"


class Video(AbcVideo):
    "動画データです。"

    rating: Rating
    "動画のレートです。"
    isPrivate: bool
    "プライベートかどうかです。"
    isDeleted: bool
    "削除済みかどうかです。"
    isNoBanner: bool
    "不明"
    isAuthenticationRequired: bool
    "不明"
    isEmbedPlayerAllowed: bool
    "埋め込み動画プレイヤーが有効かどうかです。"
    isGiftAllowed: bool
    "ギフトが有効かどうかです。"
    viewer: Viewer
    "クライアントのデータです。"
    watchableUserTypeForPayment: str # TODO: ここに入る文字列は決まっている、なのでLiteralにしたい。
    "不明"
    commentableUserTypeForPayment: str # TODO: 上記と同じ。
    "不明"


class CommentServer(DictFromAttribute):
    "コメントサーバーに関するデータです。"

    url: str
    "コメントサーバーのURLです。"


class NvCommentID(DictFromAttribute):
    "コメントデータ取得に関するデータです。"

    id: str
    fork: str


class NvCommentParams(DictFromAttribute):
    "コメントデータ取得に関するデータです。"

    targets: List[NvCommentID]
    language: str

class NvComment(DictFromAttribute):
    "コメントデータ取得に関するデータです。"

    threadKey: str
    server: str
    params: NvCommentParams


class Comment(DictFromAttribute):
    "コメントに関するデータです。"

    server: CommentServer
    nvComment: NvComment


class MyListOwner(AbcOwner):
    "マイリストのオーナーのデータです。"

    ownerType: str
    "不明"


class MyListItemVideo(AbcVideo):
    "マイリストのアイテムの動画です。"

    type: str # TODO: ここに入る文字列は恐らく決まっている...? Literalにしたい。
    "不明"
    shortDescription: str
    "短くした動画の説明です。"
    latestCommentSummary: str
    "最新のコメントの内容です。"
    isChannelVideo: bool
    "不明"
    isPaymentRequired: bool
    "不明"
    playbackPosition: Any # 何なのか調べる。
    "不明"
    owner: MyListOwner
    "マイリストの作成者のデータです。"
    requireSensitiveMasking: bool
    "センシティブなもので一度隠す必要があるかどうかです。"
    videoLive: Any # TODO: なんなのか調べる。
    "不明"

    def get_video(self) -> RealVideo:
        ":class:`niconico.video.Video` のインスタンスを作ります。"
        return self.__super__.get_video(self.url)

    @property
    def url(self) -> str:
        "この動画のURLです。"
        return f"https://www.nicovideo.jp/watch/{self.id}"


class MyListItem(DictFromAttribute["VideoClient"]):
    "マイリストのアイテムです。"

    __extends__ = {
        "video": MyListItemVideo
    }

    itemId: int
    "アイテムIDです。"
    watchId: str
    "アイテムの動画IDです。"
    description: str
    "アイテムの説明です。"
    addedAt: str
    "マイリストに追加された時間です。"
    status: str # TODO: ここに入っているのは決まっていると思うのでLiteralにしたい。
    "公開されているなら`public`が入ります。"
    video: MyListItemVideo
    "動画情報のクラスです。"


class MyList(DictFromAttribute["VideoClient"]):
    "一ページ単位でのマイリストです。"

    __extends__ = {
        "items": MyListItem
    }

    id: str
    "マイリストのIDです。"
    name: str
    "マイリストの名前です。"
    description: str
    "マイリストの説明です。"
    defaultSortKey: str
    "不明"
    defaultSortOrder: str
    "不明"
    items: list[MyListItem]
    "マイリストに入っている動画のアイテムのリストです。"
    totalItemCount: int
    "マイリストにあるアイテムの数です。"
    hasNext: bool
    "分割されたマイリストのアイテムリストのページで次のページがあるかどうかです。"
    isPublic: bool
    "公開されているかどうかです。"
    owner: MyListOwner
    "マイリストの作成者の情報です。"
    hasInvisibleItems: bool
    "不明"
    followerCount: int
    "フォロワーの数です。"
    isFollowing: bool
    "フォローしているかどうかです。"


# niconico.py - Exceptions


__all__ = ("ExtractFailed","LoginFailureException")


class ExtractFailed(Exception):
    "情報取り出しに失敗した際に発生するエラーです。"
    pass

class LoginFailureException(Exception):
    """ログインの失敗を知らせる例外クラスです。"""
    pass


# niconico.py - Search
from typing import List, TYPE_CHECKING

if TYPE_CHECKING:
    from niconico import NicoNico


class SearchClient:
    "スナップショットAPIを利用するためのクライアントです"
    APIURL = "https://api.search.nicovideo.jp/api/v2/snapshot/video/contents/search"
    def __init__(self, client):
        self.__client = client
        
    def search(self, keyword: str, *,
               sort: str="-viewCounter", targets: List[str]=["title"],
              fields: List[str]=["contentId", "title", "viewCounter"]):
        """ニコニコ動画の動画をスナップショットAPIを利用して検索します。
        Parameters
        ----------
        keyword : str
            検索キーワード
        sort : str
            ソート方式
        targets : List[str]
        fields : List[str]"""
        query = {
            "q": keyword,
            "_sort": sort,
            "targets": ",".join(target for target in targets),
            "fields": ",".join(field for field in fields)
        }
        return self.__client.request("GET", self.APIURL, params=query).json()
